/****************************************************************************************/
/*											*/
/* This program is free software; you can redistribute it and/or modify it under the	*/
/* terms of the GNU General Public License as published by the Free Software		*/
/* Foundation; either version 2 of the License, or (at your option) any later version.	*/
/*											*/
/* This program is distributed in the hope that it will be useful, but WITHOUT ANY	*/
/* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A	*/
/* PARTICULAR PURPOSE. See the GNU General Public License for more details.		*/
/*											*/
/* You should have received a copy of the GNU General Public License along with this	*/
/* program; (See "LICENSE.GPL"). If not, write to the Free Software Foundation, Inc.,	*/
/* 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.				*/
/*											*/
/****************************************************************************************/
/*											*/
/*		Leon Vinken, The Netherlands						*/
/*		leon.vinken@hetnet.nl							*/
/*											*/
/****************************************************************************************/

// Import MusicXML


// LVIFIX: add error handling/reporting for beams

// LVIFIX: check handling of key signature and accidentals

#include "resource.h"			// must be before musicxmlimport.h
#include "musicxmlimport.h"

#if GCC_MAJ_VERS > 2
#include <iostream>
#else
#include <iostream.h>
#endif
using namespace std;

#include "chord.h"
#include "muselement.h"
#include "parsertypes.h"
#include "rest.h"
#include "staff.h"
#include "voice.h"

// LVIFIX: move this (class variable ? member of struct pending ?)
static NMusElement *beamstart_chord;

bool MusicXMLParser::readStaffs(const char *fname, QList<NVoice> *voilist, QList<NStaff> *stafflist, NMainFrameWidget *mainWidget) {
//	FILE *fp;
	int i;
	int parser_return;
	NVoice *voice_elem;
	NStaff *staff_elem;
	QList<NVoice> newVoices;
	QList<NStaff> newStaffs;
	layoutDef *layoutinfo;
	int staffCount;

	parser_params.fname = fname;
	parser_params.mainWidget = mainWidget;
	parser_params.newStaffs = &newStaffs;
	parser_params.newVoices = &newVoices;

	QString strFname(fname);
	// LVIFIX: error handling
	cout << "NFileHandler::readStaffs reading XML file '" << strFname << "'" << endl;
	init_parser_variables();
	parser_return = parseMusicXML();
	cleanup_parser_variables();
	if (parser_return != 0) {
		while (!newVoices.isEmpty()) {
			newVoices.first();
			newVoices.current()->emptyVoice();
			newVoices.remove();
		}
		return false;
	}
		
	while (!voilist->isEmpty()) {
		voilist->first();
		voilist->current()->emptyVoice();
		voilist->remove();
	}
	for (voice_elem = newVoices.first(); voice_elem; voice_elem = newVoices.next()) {
		voilist->append(voice_elem);
	}
	stafflist->clear();
	for (staff_elem = newStaffs.first(); staff_elem; staff_elem = newStaffs.next()) {
		stafflist->append(staff_elem);
	}
	mainWidget->scTitle_ = parser_params.scTitle_;
	mainWidget->scSubtitle_ = parser_params.scSubtitle_;
	mainWidget->scAuthor_ = parser_params.scAuthor_;
	mainWidget->scLastAuthor_ = parser_params.scLastAuthor_;
	mainWidget->scCopyright_ = parser_params.scCopyright_;
	mainWidget->scComment_ = parser_params.scComment_;
	mainWidget->setParamsEnabled(parser_params.enableParams);
	mainWidget->setSaveWidth(parser_params.paperwidth);
	mainWidget->setSaveHeight(parser_params.paperheight);
	mainWidget->setWithMeasureNums(parser_params.with_measnum);
	delete mainWidget->braceMatrix_;
	delete mainWidget->bracketMatrix_;
	delete mainWidget->barCont_;
	staffCount = stafflist->count();
	mainWidget->braceMatrix_ = new layoutDef[staffCount];
	mainWidget->bracketMatrix_ = new layoutDef[staffCount];
	mainWidget->barCont_ = new layoutDef[staffCount];
	i = 0;
	for (layoutinfo = parser_params.bracketList.first(); i < staffCount && layoutinfo; layoutinfo = parser_params.bracketList.next()) {
		mainWidget->bracketMatrix_[i++] = *layoutinfo;
	}
	i = 0;
	for (layoutinfo = parser_params.braceList.first(); i < staffCount && layoutinfo; layoutinfo = parser_params.braceList.next()) {
		mainWidget->braceMatrix_[i++] = *layoutinfo;
	}
	i = 0;
	for (layoutinfo = parser_params.contList.first(); i < staffCount && layoutinfo; layoutinfo = parser_params.contList.next()) {
		mainWidget->barCont_[i++] = *layoutinfo;
	}
	return true;
}
// parseMusicXML(): the public interface to the MusicXML parser
//
// in: parser_params.fname contains the name of the file to be parsed
//
// out:
// on succes: returns 0
// at least one staff and one voice are created
// on failure: returns non-zero
// no staffs are allocated (voices are cleaned up by the caller)

int MusicXMLParser::parseMusicXML() {
	QString strFname(parser_params.fname);
	cout << "parseMusicXML reading XML file '" << strFname << "'" << endl;
	// LVIFIX: error handling (opening/reading file, parsing, etc)
	MusicXMLParser handler;
	QFile xmlFile(strFname);
	QXmlInputSource source(xmlFile);
	QXmlSimpleReader reader;
	reader.setContentHandler(&handler);
	cout << "parseMusicXML parsing XML file ..." << endl;
	// LVIFIX: reader.parse returns OK if the input file doesn't exist
	bool res = reader.parse(source);
	cout << "parseMusicXML done, result=" << res << endl;
	if (res) {
		return 0;
	}
	return 1;
}


// helpers for the parser

void MusicXMLParser::reportError(const QString& err)
{
	QString filename(parser_params.fname);
	QString fullErr;
	QString linenr;
	linenr.setNum(lctr->lineNumber());
	fullErr  = "\nError: In ";
	fullErr += filename;
	fullErr += " line ";
	fullErr += linenr;
	fullErr += ": ";
	fullErr += err;
	fullErr += "\n";
	cerr << fullErr;
	// LVIFIX: in future, (probably) use:
	// NResource::printWarning(fullErr);
}


// convert MusicXML note type to NoteEdit notelength

int MusicXMLParser::mxmlNoteType2Ne(const QString& mxmlNoteType)
{
	if (mxmlNoteType == "breve") {
		return DOUBLE_WHOLE_LENGTH;
	} else if (mxmlNoteType == "whole") {
		return WHOLE_LENGTH;
	} else if (mxmlNoteType == "half") {
		return HALF_LENGTH;
	} else if (mxmlNoteType == "quarter") {
		return QUARTER_LENGTH;
	} else if (mxmlNoteType == "eighth") {
		return NOTE8_LENGTH;
	} else if (mxmlNoteType == "16th") {
		return NOTE16_LENGTH;
	} else if (mxmlNoteType == "32nd") {
		return NOTE32_LENGTH;
	} else if (mxmlNoteType == "64th") {
		return NOTE64_LENGTH;
	} else if (mxmlNoteType == "128th") {
		return NOTE128_LENGTH;
	} else {
		return 0;
	}
}


// the actual parser itself

MusicXMLParser::MusicXMLParser()
	: QXmlDefaultHandler()
{
//	cout << "MusicXMLParser::MusicXMLParser" << endl;
	current_staff = 0;
	current_voice = 0;
}


// start of document handler

bool MusicXMLParser::startDocument()
{
	// LVIFIX: clearing copyright should be done by init_parser_variables
	parser_params.scCopyright_ = "";
	// init global variables: attributes, LVIFIX default to 4/4
	iDiv = 0;
	// init global variables: clear part list
	partIds.clear();
	// init global variables: characters collected
	stCha = "";
	// init global variables: identification
	stCrt = "";
	stEnc = "";
	stRig = "";
	stTtl = "";
	// LVIFIX: move this ?
	beamstart_chord = NULL;
	return TRUE;
}

// set document locator handler

void MusicXMLParser::setDocumentLocator(QXmlLocator *locator)
{
	lctr = locator;
}

// start of element handler

bool MusicXMLParser::startElement( const QString&, const QString&, 
                                   const QString& qName, 
                                   const QXmlAttributes& attributes)
{
	if (false) {
	} else if (qName == "attributes") {
    		// re-init attribute specific variables
		initStAttributes();
	} else if (qName == "beam") {
    		// store the beam number
		stBnr = attributes.value("number");
	} else if (qName == "clef") {
    		// store the clef number
		stCln = attributes.value("number");
	} else if (qName == "grace") {
		stGsl = (attributes.value("slash") == "yes");
	} else if (qName == "lyric") {
    		// store the lyric number
		stLyn = attributes.value("number");
	} else if (qName == "note") {
    		// re-init note specific variables
		initStNote();
	} else if (qName == "part") {
		// start of staff data found
		// use part id to switch to correct staff
		QString id = attributes.value("id");
		int index = -1;
		for (unsigned int i = 0; i < partIds.size(); i++) {
			if (id.compare(*partIds.at(i)) == 0) {
				index = i;
			}
		}
		if (index == -1) {
			// part id not found
			current_staff = NULL;
			current_2ndstaff = NULL;
			current_voice = NULL;
			first_voice_2ndstaff = 0;
//			cout << "MusicXMLParser::startElement(\"part\")"
//				<< " '" << id << "' not found" << endl;
		} else {
			// init vars for staff reading
			current_staff = parser_params.newStaffs->at(index);
			current_2ndstaff = NULL;
			current_voice = current_staff->getVoiceNr(0);
			first_voice_2ndstaff = 0;
//			cout << "MusicXMLParser::startElement(\"part\")"
//				<< " '" << id << "' is staff " << index << endl;
		}
		current_time = 0;
	} else if (qName == "score-part") {
		// start of staff definition found
		// re-init score part specific variables
		initStScorePart();
		stPid = attributes.value("id");
	} else if (qName == "tuplet") {
    		// store the tuplet type
		stTtp = attributes.value("type");
	} else {
		// others (silently) ignored
	}
	return TRUE;
}


// end of element handler

bool MusicXMLParser::endElement( const QString&, const QString&,
                                  const QString& qName)
{
	QString Str;
	if (false) {
	} else if (qName == "accidental") {
		stAcc = stCha;
	} else if (qName == "actual-notes") {
		stAno = stCha;
	} else if (qName == "alter") {
		stAlt = stCha;
	} else if (qName == "attributes") {
		handleAttributes();
		initStAttributes();
	} else if (qName == "backup") {
		// LVIFIX: error handling
//		cout << "--> MusicXMLParser::endElement() backup "
//			<< stDur << endl;
		current_time -= stDur.toInt();
	} else if (qName == "beam") {
		// LVIFIX: keep state only for beam level 1
		if (stBnr == "1") {
			stBea = stCha;
		}
	} else if (qName == "beats") {
		stBts = stCha;
	} else if (qName == "beat-type") {
		stBtt = stCha;
	} else if (qName == "chord") {
		stCho = TRUE;
	} else if (qName == "clef-octave-change") {
		if (stCln == "2") { stCoc2 = stCha; } else { stCoc = stCha; }
	} else if (qName == "creator") {
		stCrt = stCha;
	} else if (qName == "divisions") {
		stDiv = stCha;
		iDiv = stDiv.toInt();
		if (iDiv <= 0) {
			Str = "illegal <divisions> value: " + stDiv;
			reportError(Str);
		}
	} else if (qName == "dot") {
		stDts++;
	} else if (qName == "duration") {
		stDur = stCha;
	} else if (qName == "encoder") {
		stEnc = stCha;
	} else if (qName == "fifths") {
		stFif = stCha;
	} else if (qName == "forward") {
		// LVIFIX: error handling
//		cout << "--> MusicXMLParser::endElement() forward "
//			<< stDur << endl;
		current_time += stDur.toInt();
	} else if (qName == "grace") {
		stGra = true;
	} else if (qName == "identification") {
		// LVIFIX: remove_newlines necessary ?
		parser_params.scAuthor_    = stCrt;
		parser_params.scCopyright_ = stRig;
		parser_params.scTitle_     = stTtl;
//		ts->transcriber = stEnc;
//		ts->comments    = "";
	} else if (qName == "line") {
		if (stCln == "2") { stCli2 = stCha; } else { stCli = stCha; }
	} else if (qName == "measure") {
		NSign *sign;
		NStaff *staff;
		NVoice *voice;
		staff = current_staff;
		staff->actualKeysig_.resetAtBar();
		sign = new NSign(&(parser_params.mainWidget->main_props_),
				 &(staff->staff_props_), SIMPLE_BAR);
		voice = staff->getVoiceNr(0);
		voice->appendElem(sign);
//		cout << "--> MusicXMLParser::endElement() measure,"
//			<< " simple bar @" << sign << endl;
		if (current_2ndstaff) {
			staff = current_2ndstaff;
			staff->actualKeysig_.resetAtBar();
			sign = new NSign(&(parser_params.mainWidget->main_props_),
					 &(staff->staff_props_), SIMPLE_BAR);
			voice = staff->getVoiceNr(0);
			voice->appendElem(sign);
//			cout << "--> MusicXMLParser::endElement() measure,"
//				<< " simple bar @" << sign << endl;
		}
//	} else if (qName == "midi-bank") {
//		stPmb = stCha;
	} else if (qName == "midi-channel") {
		stPmc = stCha;
	} else if (qName == "midi-program") {
		stPmp = stCha;
	} else if (qName == "normal-notes") {
	    stNno = stCha;
	} else if (qName == "note") {
		return addNote();
	} else if (qName == "octave") {
		stOct = stCha;
	} else if (qName == "part-name") {
		stPnm = stCha;
	} else if (qName == "rest") {
	    stRst = true;
	} else if (qName == "rights") {
		stRig = stCha;
	} else if (qName == "score-part") {
		bool res = addStaff();
		// re-init score part specific variables
		initStScorePart();
		return res;
	} else if (qName == "sign") {
		if (stCln == "2") { stCsi2 = stCha; } else { stCsi = stCha; }
	} else if (qName == "staves") {
		stSta = stCha;
//		cout << "--> MusicXMLParser::endElement() staves "
//			<< stSta << endl;
		if (stSta == "2") {
			addSecondStaff();
		}
	} else if (qName == "staff") {
		stStf = stCha;
	} else if (qName == "stem") {
		stStm = stCha;
	} else if (qName == "step") {
		stStp = stCha;
	} else if (qName == "text") {
		stTxt = stCha;
		handleLyrics();
	} else if (qName == "type") {
		stTyp = stCha;
	} else if (qName == "voice") {
		stVoi = stCha;
	} else if (qName == "work-title") {
		stTtl = stCha;
	// following elements are explicitly ignored, usually because sub-
	// and superelements handle all the work but sometimes they are
	// simply not necessary
	} else if (
			   qName == "clef"
			|| qName == "encoding"
			|| qName == "instrument-name"
			|| qName == "key"
			|| qName == "lyric"
			|| qName == "midi-instrument"
			|| qName == "mode"
			|| qName == "part-list"
			|| qName == "part"
			|| qName == "pitch"
			|| qName == "score-instrument"
			|| qName == "score-partwise"
			|| qName == "software"
			|| qName == "syllabic"
			|| qName == "time"
			|| qName == "work"
//			|| qName == ""
		  ) {
	} else {
		// LVIFIX: define what to do, e.g. silently ignore unknown
		// elements, or report these as warning
		// for the time being, the last option is chosen
		cerr << "--> MusicXMLParser::endElement() skipping '<"
			<< qName << ">'" << endl;
	}
	return TRUE;
}


// character(s) handler

bool MusicXMLParser::characters(const QString& ch)
{
	stCha = ch;
	return TRUE;
}


// add a note to the current track
// LVIFIX: error handling: int conversion, limits
// LVIFIX: compare current_time with amount of data in the voice
// if current_time is smaller, report error
// if equal, do nothing
// if larger, insert rests

bool MusicXMLParser::addNote()
{
/*
	cout << "MusicXMLParser::addNote()"
		<< " acc="   << stAcc
		<< " alter=" << stAlt
		<< " chord=" << stCho
		<< " dots="  << stDts
		<< " durat=" << stDur
		<< " octav=" << stOct
		<< " rest="  << stRst
		<< " staff=" << stStf
		<< " stem="  << stStm
		<< " step="  << stStp
		<< " type="  << stTyp
		<< " voice=" << stVoi
		<< endl;
	cout << " lyrics:";
	for (int i = 0; i < NUM_LYRICS; i++) {
		cout << " '" << stLyr[i] << "'";
	}
	cout
		<< " ano="  << stAno
		<< " nno="  << stNno
		<< " ttp="  << stTtp
		<< " gra="  << stGra
		<< " grsl=" << stGsl
		<< endl;
	cout << " current_time(before)=" << current_time;
*/
	int dur;
	int line = 0;
	int length = mxmlNoteType2Ne(stTyp);
	int offs = 0;
	bool res;
	unsigned int status = 0;
	QString Str;
	if ((stTyp != "") && (length == 0)) {
		Str = "illegal <type> value: " + stTyp;
		reportError(Str);
	}
	if (stGra) {
		dur = 0;
		if (stTyp == "eighth") {
			status = STAT_GRACE;
			if (stGsl) {
				length = INTERNAL_MARKER_OF_STROKEN_GRACE;
			}
		} else if (stTyp == "16th") {
			status = STAT_GRACE;
		} else {
			Str = "illegal grace note <type>: " + stDur;
			reportError(Str);
		}
	} else {
		dur = stDur.toInt(&res);
		if (!res || (dur < 0)) {
			dur = 0;		// sensible default (?)
			Str = "illegal <duration> value: " + stDur;
			reportError(Str);
		}
	}
	if (stDts == 0) {
		/* do nothing status = 0 */ ;
	} else if (stDts == 1) {
		status |= STAT_SINGLE_DOT;
	} else if (stDts == 2) {
		status |= STAT_DOUBLE_DOT;
	} else {
		Str.setNum(stDts);
		Str = "illegal number of dots: " + Str;
		reportError(Str);
	}
	if (stAcc == "") {
		offs = UNDEFINED_OFFS;
	} else if (stAcc == "flat") {
		offs = -1; status |= STAT_FORCE;
	} else if (stAcc == "natural") {
		offs =  0; status |= STAT_FORCE;
	} else if (stAcc == "sharp") {
		offs =  1; status |= STAT_FORCE;
	} else {
		Str = "illegal <accidental> value: " + stAcc;
		reportError(Str);
	}
	int stemPol = STEM_POL_INDIVIDUAL;
	if (stStm == "") {
		/* ignore */;
	} else if (stStm == "up") {
		stemPol = STEM_POL_UP;
	} else if (stStm == "down") {
		stemPol = STEM_POL_DOWN;
	} else {
		Str = "illegal <stem> dir: " + stStm;
		reportError(Str);
	}
//	cout << endl << " ";
	NStaff *staff = NULL;
	if ((stStf == "") || (stStf == "1")) {
		staff = current_staff;
	} else if (stStf == "2") {
		staff = current_2ndstaff;
	} else {
		Str = "illegal <staff>: " + stStf;
		reportError(Str);
	}
	if (stVoi == "") {
		// default to voice 1
		stVoi = "1";
	}
	/*
	if ((stVoi != "1") && (stVoi != "2")) {
		Str = "illegal <voice>: " + stVoi;
		reportError(Str);
	}
	*/
	handleVoice(stStf.toInt(), stVoi.toInt());
//	cout << " ";
	setCurrentTime();
	// LVIFIX: beam error handling is essentially missing
	bool beamStart = ((stBea == "begin"));
	bool beamStop  = ((stBea == "end"));
//	cout << " beamStart=" << beamStart
//		<< " beamStop=" << beamStop;
	if (stRst) {
		if (stTyp == "") {
			// <type> missing, insert rest(s) based on duration
			insertRest(dur * QUARTER_LENGTH / iDiv, false);
		} else {
		NRest * rest = new NRest(&(parser_params.mainWidget->main_props_),
					 &(staff->staff_props_),
					 &(current_voice->yRestOffs_),
					 length,
					 status);
		current_voice->appendElem(rest);
//		cout << " rest @" << rest;
		current_time += dur;
		}
	} else {
		NClef *clef = getClefAt(staff, current_time * QUARTER_LENGTH / iDiv);
/*
		cout
			<< " clef=" << clef
			<< " line=" << line;
		cout << " name2Line=" << clef->name2Line(*(stStp.lower().latin1()))
			<< " getOctave=" << clef->getOctave()
			<< " clefKind=" << clef->getSubType();
*/
		line = clef->name2Line(*(stStp.lower().latin1()));
		line += 7 * (stOct.toInt() - clef->getOctave());
		// LVIFIX: tenor clef needs one octave correction,
		// apparently caused by name2line
		if (clef->getSubType() == TENOR_CLEF) {
			line += 7;
		}
		if (stCho) {
			NMusElement *elem = current_voice->getLast();
//			cout << " <chord/>, lastPos=" << elem;
			if (elem && (elem->getType() == T_CHORD)) {
				NChord * chrd = (NChord *) elem;
				(void) chrd->insertNewNote(line,
							   offs,
							   stemPol,
							   status);
			} else {
				Str = "<chord/> without preceding note";
				reportError(Str);
			}
		} else {
			NChord * chrd = new NChord(&(parser_params.mainWidget->main_props_),
						   &(staff->staff_props_),
					           line,
						   offs,
						   length,
						   stemPol,
						   status);
			current_voice->appendElem(chrd);
//			cout << " chord @" << chrd;
			current_time += dur;
			if (beamStart) {
				beamstart_chord = chrd;
			}
			if (beamStop && beamstart_chord) {
				if (!current_voice->buildBeam(beamstart_chord, chrd)) {
					Str = "could not build beam";
					reportError(Str);
				}
			}
			for (int i = 0; i < NUM_LYRICS; i++) {
				if (stLyr[i] != "") {
					chrd->setLyrics(&stLyr[i], i);
				}
			}
			// note: MusicXML adds tuplet info to each note of a
			// chord, while in NoteEdit tuplet info is associated 
			// with the chord itself
			handleTuplet(chrd);
		}
	}
//	cout << " current_time(after)=" << current_time
//		<< endl;
	// re-init note specific variables
	initStNote();
	return TRUE;
}


// add a staff to the current song

bool MusicXMLParser::addStaff()
{
	bool res;
	QString Str;
	int i = parser_params.newStaffs->count();
	// new staff found, append it
/*
	cout << "MusicXMLParser::addStaff()"
		<< " nStaffs=" << i
		<< " stPid=" << stPid
//		<< " stPmb=" << stPmb
		<< " stPmc=" << stPmc
		<< " stPmp=" << stPmp
		<< " stPnm='" << stPnm << "'";
*/
	// create staff and first voice
	parser_params.newStaffs->append(
		current_staff = new NStaff(i*(NResource::overlength_+STAFF_HIGHT
					+NResource::underlength_)
					+NResource::overlength_,
					i, 0, parser_params.mainWidget));
	parser_params.newVoices->append(current_staff->getVoiceNr(0));
	// remember part id to staff nr mapping by storing the part id's
	QString *sp = new QString(stPid);
	int sz = partIds.size();
	partIds.resize(sz+1);
	partIds.insert(sz, sp);
	// set staff name
	current_staff->staffName_ = stPnm;	// LVIFIX: check UTF-8 correct
	// set staff's MIDI channel
	int channel = stPmc.toInt(&res);
	if (res) {
		if (channel < 1 || channel > 16) {
			res = false;
		} else {
			current_staff->setChannel(channel - 1);
		}
	}
	if (!res) {
		Str = "bad channel: " + stPmc;
		reportError(Str);
	}

	// LVIFIX: setting current_voice should probably (also ?) be done
	// in something like select_voice_and_staff(...)
	current_voice = parser_params.newVoices->first();
//	cout
//		<< " current_staff=" << current_staff
//		<< " current_voice=" << current_voice
//		<< endl;

	// set staff's MIDI program
	int program = stPmp.toInt(&res);
	if (res) {
		if (program < 1 || program > 128) {
			res = false;
		} else {
			current_staff->changeVoice(program - 1);
			current_voice->voiceSet_ = true;
		}
	}
	if (!res) {
		Str = "bad voice: " + stPmp;
		reportError(Str);
	}

	return TRUE;
}


// add a second staff to the current one
// this implies:
// - adding a new staff to parser_params.newStaffs after the current staff
//   with relevant parameters copied from the current staff
// - adding "" to partIds at the corresponding position
// - TBD initializing variables for the second staff
//   may need to be done at start of part

bool MusicXMLParser::addSecondStaff()
{
//	cout << "MusicXMLParser::addSecondStaff()"
//		<< " current_staff=" << current_staff
//		<< " newStaffs->at()=" << parser_params.newStaffs->at();
	// add a new staff to parser_params.newStaffs after the current staff
	int i = parser_params.newStaffs->at() + 1;
	parser_params.newStaffs->insert(i,
		current_2ndstaff = new NStaff(i*(NResource::overlength_+STAFF_HIGHT
					+NResource::underlength_)
					+NResource::overlength_,
					i, 0, parser_params.mainWidget));
	parser_params.newVoices->append(current_2ndstaff->getVoiceNr(0));
	// remember part id to staff nr mapping by storing the part id's
	QString *sp = new QString("");
	int sz = partIds.size();
	partIds.resize(sz+1);
	for (int j = sz - 1; j >= i; j--) {
//		cout << " " << j << ">" << j + 1;
		partIds.insert(j + 1, partIds.at(j));
	}
	partIds.insert(i, sp);
//	cout
//		<< " current_2ndstaff=" << current_2ndstaff
//		<< endl;
	// copy relevant parameters from the current staff:
	// staff name
	current_2ndstaff->staffName_ = current_staff->staffName_;	// LVIFIX: check UTF-8 correct
	// staff's MIDI channel
	current_2ndstaff->setChannel(current_staff->getChannel());
	// staff's MIDI program
	current_2ndstaff->changeVoice(current_staff->getVoice());
	current_2ndstaff->getVoiceNr(0)->voiceSet_ = true;
	// debug: dump data
/*
	int oldat = parser_params.newStaffs->at();
	cout << "partIds/newStaffs=";
	for (unsigned int u = 0; u < parser_params.newStaffs->count(); u++) {
		cout
			<< " " << (*partIds.at(u))
			<< "/" << parser_params.newStaffs->at(u);
	}
	parser_params.newStaffs->at(oldat);
	cout << endl;
*/
	return TRUE;
}

// get the clef that is in effect in staff at miditime
// return NResource::nullClef_ if no clef found

NClef * MusicXMLParser::getClefAt(NStaff *staff, int miditime)
{
	NClef *clef = NResource::nullClef_;
	NVoice * voice_elem = staff->getVoiceNr(0);
	voice_elem->prepareForWriting();
	NMusElement *elem;
	elem = voice_elem->getCurrentPosition();
	if (elem == 0) return clef;
	do {
		if ((elem->getType() == T_CLEF) && (elem->midiTime_ <= miditime)) {
			clef = (NClef *) elem;
		}
		elem = voice_elem->getNextPosition();
	} while (elem);
	return clef;
}


// handle attributes
// note that the attribute order in the MusicXML files normally differs
// from what is needed in NoteEdit (which uses the order as is printed):
// MusicXML: key, time, clef
// NoteEdit: clef, key, time
// Also, in piano music two clefs are usually (but not always !) found

void MusicXMLParser::handleAttributes()
{
/*
	cout << "MusicXMLParser::handleAttributes()"
		<< " stBts=" << stBts
		<< " stBtt=" << stBtt
		<< " stCli=" << stCli
		<< " stCoc=" << stCoc
		<< " stCsi=" << stCsi
		<< " stCli2=" << stCli2
		<< " stCoc2=" << stCoc2
		<< " stCsi2=" << stCsi2
		<< " stDiv=" << stDiv
		<< " stFif=" << stFif
		<< endl;
*/
	bool res;
	QString Str;
	// handle the clef(s)
	handleClef(current_staff, stCli, stCoc, stCsi);
	handleClef(current_2ndstaff, stCli2, stCoc2, stCsi2);
	// handle the key signature
	// if stFif is empty, assume no key signature
	int iFif = 0;
	if (stFif != "") {
		iFif = stFif.toInt(&res);
		if (res) {
			if ((-6 <= iFif) && (iFif <= 6)) {
				int count;
				int kind;
				if (iFif < 0) {
					count = -iFif; kind = STAT_FLAT;
				} else {
					count =  iFif; kind = STAT_CROSS;
				}
				if (count > 0) {
					// LVIFIX: check handling accidentals and key signature
					/*
					current_staff->actualKeysig_.setRegular(count, kind);
					current_staff->actualKeysig_.setClef(&(current_staff->actualClef_));
					NKeySig * ksig = new NKeySig(current_voice->getMainPropsAddr(),
								     current_staff->getStaffPropsAddr());
					ksig->change(&(current_staff->actualKeysig_));
					current_voice = current_staff->getVoiceNr(0);
					current_voice->appendElem(ksig);
					*/
					NStaff *staff = 0;
					NVoice *voice = 0;
					staff = current_staff;
					voice = staff->getVoiceNr(0);
					NClef *clef = getClefAt(staff, current_time * QUARTER_LENGTH / iDiv);
					NKeySig * ksig = new NKeySig(voice->getMainPropsAddr(),
								     staff->getStaffPropsAddr());
					ksig->setRegular(count, kind);
					ksig->setClef(clef);
					voice->appendElem(ksig);
//					cout << " ksig @" << ksig
//						<< " w/ clef @" << clef << endl;
					if (current_2ndstaff) {
						staff = current_2ndstaff;
						voice = staff->getVoiceNr(0);
						NClef *clef = getClefAt(staff, current_time * QUARTER_LENGTH / iDiv);
						NKeySig * ksig = new NKeySig(voice->getMainPropsAddr(),
									     staff->getStaffPropsAddr());
						ksig->setRegular(count, kind);
						ksig->setClef(clef);
						voice->appendElem(ksig);
//						cout << " ksig @" << ksig
//							<< " w/ clef @" << clef << endl;
					}
				}
			} else {
				res = false;
			}
		}
		if (!res) {
			Str = "illegal <fifths> value: " + stFif;
			reportError(Str);
		}
	}
	// handle the time signature
	// if both stBts and stBtt are empty, assume no time signature
	if ((stBts != "") && (stBtt != "")) {
		int iBts = 0;
		int iBtt = 0;
		bool res1 = true;
		bool res2 = true;
		iBts = stBts.toInt(&res1);
		iBtt = stBtt.toInt(&res1);
		// LVIFIX: improve error handling
		if (res1 && res2) {
			NVoice *voice;
			current_staff->staff_props_.measureLength = iBts * 128 / iBtt;
			voice = current_staff->getVoiceNr(0);
			voice->appendElem(T_TIMESIG, iBts, iBtt);
			if (current_2ndstaff) {
				voice = current_2ndstaff->getVoiceNr(0);
				voice->appendElem(T_TIMESIG, iBts, iBtt);
			}
		} else {
			Str = "illegal <time> value: " + stBts + "/" + stBtt;
			reportError(Str);
		}
	}
}


// append a clef to staff

void MusicXMLParser::handleClef(NStaff *staff, QString& cli, QString& coc, QString& csi)
{
	if (staff == NULL) return;
	if ((cli == "") && (coc == "") && (csi == "")) return;
	bool res = true;
	QString Str;
	// handle the clef
	// if both csi and cli are empty, assume no clef
	int kind = 0;
	int iCoc = 0;
	if (res) {
		if ((csi == "G") && (cli == "2")) {
			kind = TREBLE_CLEF;
		} else if ((csi == "F") && (cli == "4")) {
			kind = BASS_CLEF;
		} else if ((csi == "C") && (cli == "3")) {
			kind = ALTO_CLEF;
		} else if ((csi == "C") && (cli == "4")) {
			kind = TENOR_CLEF;
		} else {
			Str = "bad clef: ";
			Str += csi;
			Str += cli;
			reportError(Str);
			res = false;
		}
	}
	if (res) {
		if (coc == "-1") {
			iCoc = -8;
		} else if (coc == "") {
			iCoc =  0;
		} else if (coc == "1") {
			iCoc =  8;
		} else {
			Str = "bad octave shift: ";
			Str += coc;
			reportError(Str);
			res = false;
		}
	}
	if (res) {
		NClef *c;
		NVoice *voice = staff->getVoiceNr(0);
		c = new NClef(voice->getMainPropsAddr(),
				staff->getStaffPropsAddr(),
				kind, iCoc);
		voice->appendElem(c);
//		cout << " clef @" << c << endl;
	}
}


// handle tuplets

// LVIFIX: move and init at start of part
static NMusElement * firstElem = 0;

void MusicXMLParser::handleTuplet(NMusElement * elem)
{
	// first check if chord is part of tuplet
	if ((stAno == "") && (stNno == "") && (stTtp == "")) {
		return;
	}
	QString Str;
	// then check error conditions
	if (stAno != "3") {
		Str = "illegal <actual-notes> value: " + stAno;
		reportError(Str);
		return;
	}
	if (stNno != "2") {
		Str = "illegal <normal-notes> value: " + stAno;
		reportError(Str);
		return;
	}
	if ((stTtp != "start") && (stTtp != "stop") && (stTtp != "")) {
		Str = "illegal <actual> type: " + stTtp;
		reportError(Str);
		return;
	}
	if (stTtp == "start") {
		firstElem = elem;
	} else if (stTtp == "stop") {
		if (firstElem == 0) {
			Str = "tuplet stop without start";
			reportError(Str);
			return;
		}
		if (!current_voice->buildTuplet(firstElem, elem, 3, 2)) {
			Str = "could not build tuplet";
			reportError(Str);
		}
		firstElem = 0;
	}
}


// handle lyrics: add a line to the stLyr array.
// in: index in stLyn (lyrics number), text in stTxt.

void MusicXMLParser::handleLyrics()
{
	int i = stLyn.toInt();
	QString Str;
	// check error conditions first
	// note that if stLyn is not an integer, i will be 0, which is caught
	if ((i <= 0) || (i > NUM_LYRICS)) {
		Str = "illegal lyrics number value: " + stLyn;
		reportError(Str);
		return;
	}
	stLyr[i - 1] = stTxt;
}

// handle staff and voice change, create voice if necessary
// note: staff_nr and voice_nr start at 1

// if staff is 0: allocate all voices to current (and only) staff
// if staff is 1:
//   if no second staff yet or voice_nr lower than start of second staff:
//     allocate on first staff
//   else error
// if staff is 2:
//   if voice_nr ligher than or equal to start of second staff:
//     allocate on second staff
//   else error

void MusicXMLParser::handleVoice(int staff_nr, int voice_nr)
{
//	cout << "MusicXMLParser::handleVoice(" << staff_nr << ", "
//		<< voice_nr << ")";
	QString Str;
	// check error conditions first
	// if second staff exists, number of voices in first staff is limited
	if (current_2ndstaff && (first_voice_2ndstaff > 0)
	    && (staff_nr < 2) && (voice_nr >= first_voice_2ndstaff)) {
		Str.setNum(voice_nr);
		Str = "voice " + Str;
		Str += " already allocated to staff 2";
		reportError(Str);
	}
	// if second staff exists, voice_nr cannot be lower than first voice
	if (current_2ndstaff && (first_voice_2ndstaff > 0)
	    && (staff_nr == 2) && (voice_nr < first_voice_2ndstaff)) {
		Str.setNum(voice_nr);
		Str = "voice " + Str;
		Str += " already allocated to staff 1";
		reportError(Str);
	}
	// if staff 2 requested, it must exist
	if ((staff_nr == 2) && (!current_2ndstaff)) {
		Str = "staff 2 does not exist";
		reportError(Str);
	}
	// create new voice(s) if necessary
	int fv = 0;			// first voice
	int nv = 0;			// number of voices
	int nnv = 0;			// number of new voices
	NStaff *staff = 0;		// actual staff used
	int vnis = 0;			// voice nr in staff
	if (staff_nr < 2) {
		// handle first staff
		staff = current_staff;
		fv = 1;
		nv = staff->voiceCount();
//		cout << " current_staff->voiceCount()=" << current_staff->voiceCount();
		nnv = voice_nr - nv;
	} else {
		// handle second staff
		staff = current_2ndstaff;
		nv = staff->voiceCount();
		if (first_voice_2ndstaff == 0) {
			// found first voice of staff 2, no need to allocate
			first_voice_2ndstaff = voice_nr;
			nnv = 0;
		} else {
			nnv = voice_nr - first_voice_2ndstaff - nv + 1;
		}
		fv = first_voice_2ndstaff;
	}
	vnis = voice_nr - fv;
	if (nnv > 0) {
//		cout << " adding " << nnv << " voice(s)";
		for (; nnv > 0; nnv--) {
			staff->addVoices(1);
//			cout << " " << voice_nr - nnv
//				<< "(" << voice_nr - fv - nnv + 1 << ")";
			parser_params.newVoices->append(
				staff->getVoiceNr(voice_nr - fv - nnv + 1));
		}
	}
//	cout
//		<< " fv=" << fv
//		<< " nv=" << nv
//		<< " staff=" << staff
//		<< " vnis=" << vnis;
	if ((vnis >= staff->voiceCount()) || (vnis < 0)) {
		Str.setNum(voice_nr);
		Str = "internal error, voice " + Str;
		Str += " not possible";
		reportError(Str);
	}
	current_voice = staff->getVoiceNr(vnis);
//	cout
//		<< " staff=" << staff
//		<< " current_voice=" << current_voice
//		<< endl;
}


// initialize attribute state variables

void MusicXMLParser::initStAttributes()
{
	stBts  = "";
	stBtt  = "";
	stCli  = "";
	stCln  = "";
	stCoc  = "";
	stCsi  = "";
	stCli2 = "";
	stCoc2 = "";
	stCsi2 = "";
	stDiv  = "";
	stFif  = "";
	stSta  = "";
}


// initialize note state variables

void MusicXMLParser::initStNote()
{
	stAcc = "";
	stAlt = "";
	stAno = "";
	stBea = "";
	stBnr = "";
	stCho = false;
	stDts = 0;
	stDur = "";
	stFrt = "";
	stGls = false;
	stGra = false;
	stGsl = false;
	stHmr = false;
	stLyn = "";
	for (int i = 0; i < NUM_LYRICS; i++) {
		stLyr[i] = "";
	}
	stNno = "";
	stOct = "";
	stPlo = false;
	stRst = false;
	stStf = "";
	stStm = "";
	stStp = "";
	stStr = "";
	stTie = false;
	stTtp = "";
	stTxt = "";
	stTyp = "";
	stVoi = "";
}


// initialize part state variables

void MusicXMLParser::initStScorePart()
{
	stPid = "";
//	stPmb = "";
	stPmc = "";
	stPmp = "";
	stPnm = "";
}


// insert (hidden) rests in the current voice
// length is split in smaller parts if necessary

void MusicXMLParser::insertRest(int length, bool hidden)
{
//	cout << " MusicXMLParser::insertRest(" << length << ", "
//		<< hidden << ")" << endl;
	for (int i = QUARTER_LENGTH; i >= NOTE128_LENGTH; i /= 2) {
		while (length >= i) {
//			cout << " " << i;
			int stat = hidden ? STAT_HIDDEN : 0;
			NRest * rest = new NRest(&(parser_params.mainWidget->main_props_),
						 &(current_staff->staff_props_),
						 &(current_voice->yRestOffs_),
						 i,
						 stat);
			current_voice->appendElem(rest);
//			cout << " rest @" << rest;
			length -= i;
		}
	}
//	cout << endl;
	if (length > 0) {
		// LVIFIX: report error
	}
}


// set the actual time to the current time in the current voice
// this implies adding hidden rests when necessary
// note that the actual amount of time in voice 0 needs to be determined,
// which may differ from the time signature

// for voices != first voice this implies:
// determine voice's midi end time
// search from there in voice 0 until end-of-bar or current time reached
// fill with hidden rests
// repeat until done

void MusicXMLParser::setCurrentTime()
{
	current_voice->computeMidiTime(false);
	int cvMidiEndTime = current_voice->getMidiEndTime();
	int currentMidiTime = current_time * QUARTER_LENGTH / iDiv;
//	cout << "MusicXMLParser::setCurrentTime()"
//		<< " current_voice=" << current_voice
//		<< " midiEndTime=" << current_voice->getMidiEndTime()
//		<< " current_time=" << current_time
//		<< " (" << current_time * QUARTER_LENGTH / iDiv << ")"
//		<< endl;
	if (current_voice->isFirstVoice()) {
//		cout << " first voice:";
		if (currentMidiTime == cvMidiEndTime) {
//			cout << " ready" << endl;
		} else {
//			cout
//				<< " error, currentMidiTime != cvMidiEndTime"
//				<< endl;
			// LVIFIX report error
		}
		return;
	}
//	cout << " other voice:";
	if (currentMidiTime < cvMidiEndTime) {
//		cout << " error currentMidiTime < cvMidiEndTime" << endl;
		// LVIFIX report error
	}
	// loop over all elements in voice 0:
	//   if barsym is found with cvMidiEndTime < miditime <= currentMidiTime:
	//     insert hidden rests (miditime - cvMidiEndTime)
	//     cvMidiEndTime = miditime
	// if currentMidiTime > cvMidiEndTime:
	//     insert hidden rests (currentMidiTime - cvMidiEndTime)
	NVoice * voice0 = current_staff->getVoiceNr(0);
	if (voice0 == 0) return;
	// LVIFIX: prepareForWriting() does more than just positioning:
	// changing theStaff's actual clef and keysig, which does hurt !
	voice0->prepareForWriting();
	NMusElement *elem = 0;
	int delta = 0;
	elem = voice0->getCurrentPosition();
//	if (elem == 0) return;
	while (elem) {
//		out_ << "elem @" << elem;
//		out_ << endl;
		if (elem->getType() == T_SIGN
		    && (elem->getSubType() & BAR_SYMS)) {
			if ((cvMidiEndTime < elem->midiTime_)
			    && (elem->midiTime_ <= currentMidiTime)) {
			    	delta = elem->midiTime_ - cvMidiEndTime;
//				cout << " insert(" << delta << ")" << endl;
				insertRest(delta, true);
				cvMidiEndTime += delta;
			}
		}
		elem = voice0->getNextPosition();
	}
	delta = currentMidiTime - cvMidiEndTime;
	if (delta > 0) {
//		cout << " insert(" << delta << ")" << endl;
		insertRest(delta, true);
		cvMidiEndTime += delta;
	}
	// LVIFIX: following code won't work when clefs change
	// hack: repair the actual clef by setting it to the first clef
//	NClef *c = voice0->getFirstClef();
//	current_staff->actualClef_.changeKind(c->getSubType());
//	current_staff->actualClef_.setShift(c->getOctave());
}
