/****************************************************************************************/
/*											*/
/* This program is free software; you can redistribute it and/or modify it under the	*/
/* terms of the GNU General Public License as published by the Free Software		*/
/* Foundation; either version 2 of the License, or (at your option) any later version.	*/
/*											*/
/* This program is distributed in the hope that it will be useful, but WITHOUT ANY	*/
/* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A	*/
/* PARTICULAR PURPOSE. See the GNU General Public License for more details.		*/
/*											*/
/* You should have received a copy of the GNU General Public License along with this	*/
/* program; (See "LICENSE.GPL"). If not, write to the Free Software Foundation, Inc.,	*/
/* 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.				*/
/*											*/
/****************************************************************************************/
/*											*/
/*		Leon Vinken, The Netherlands						*/
/*		leon.vinken@hetnet.nl							*/
/*											*/
/****************************************************************************************/

#include "config.h"
#ifndef MUSICXMLIMPORT_H

#include <qlist.h>
#include <qvector.h>
#include <qxml.h>

// #include "config.h"
// #include "staff.h"
// #include "voice.h"

int parseMusicXML();

class NClef;
class NStaff;
class NVoice;
class QString;
class NMainFrameWidget;
class NMusElement;

class MusicXMLParser : public QXmlDefaultHandler
{
public:
	MusicXMLParser();
	bool readStaffs(const char *fname, QList<NVoice> *voilist, QList<NStaff> *stafflist, NMainFrameWidget *mainWidget);
	void setDocumentLocator(QXmlLocator *locator);
	bool startDocument();
	bool startElement(const QString&, const QString&, const QString&,
	                  const QXmlAttributes&);
	bool endElement(const QString&, const QString&, const QString&);
	bool characters(const QString & ch);
private:
	bool addNote();
	bool addStaff();
	bool addSecondStaff();
	NClef *getClefAt(NStaff *staff, int miditime);
	void handleAttributes();
	void handleClef(NStaff *staff, QString& cli, QString& coc, QString& csi);
	void handleLyrics();
	void handleTuplet(NMusElement * elem);
	void handleVoice(int staff, int voice);
	void initStAttributes();
	void initStNote();
	void initStScorePart();
	void insertRest(int length, bool hidden);
	int mxmlNoteType2Ne(const QString& mxmlNoteType);
	int parseMusicXML();
	void reportError(const QString& err);
	void setCurrentTime();
	NStaff *current_staff;
	NVoice *current_voice;
	NStaff *current_2ndstaff;
	int first_voice_2ndstaff;
	int current_time;		// current time in the current staff
	int iDiv;			// divisions
	QVector<QString> partIds;	// part (staff) id's
	QXmlLocator *lctr;
	// state variables for parsing
//	static const int beamlvls = 5;		// beam levels
	// attributes -- initialized in initStAttributes()
	QString stBts;				// beats
	QString stBtt;				// beat-type
	QString stCli;				// clef line
	QString stCln;				// clef number
	QString stCoc;				// clef octave change
	QString stCsi;				// clef sign
	QString stCli2;				// clef 2 line
	QString stCoc2;				// clef 2 octave change
	QString stCsi2;				// clef 2 sign
	QString stDiv;				// divisions
	QString stFif;				// fifths
	QString stSta;				// staves
	// general -- initialized in startDocument()
	QString stCha;				// characters collected
	// identification -- initialized in startDocument()
	QString stCrt;				// creator
	QString stEnc;				// encoder
	QString stRig;				// rights
	QString stTtl;				// title
	// note (including forward/backup) -- initialized in initStNote()
	QString stAcc;				// accidental
	QString stAlt;				// alter
	QString	stAno;				// actual notes
//	QString	stBea[beamlvls];		// beam
	QString	stBea;				// beam
	QString	stBnr;				// beam number
	bool    stCho;				// chord with previous note
	int     stDts;				// dots (count)
	QString stDur;				// duration
	QString stFrt;				// fret
	bool    stGls;				// glissando
	bool    stGra;				// grace
	bool    stGsl;				// grace slash
	bool    stHmr;				// hammer-on
	QString stLyn;				// lyrics number
	QString stLyr[NUM_LYRICS];		// lyrics
	QString	stNno;				// normal notes
	QString stOct;				// octave
	bool    stPlo;				// pull-off
	bool    stRst;				// rest
	QString stStf;				// staff
	QString stStm;				// stem
	QString stStp;				// step
	QString stStr;				// string
	bool    stTie;				// tie stop
	QString stTtp;				// tuplet type
	QString stTxt;				// text
	QString stTyp;				// type
	QString stVoi;				// voice
	// part (== staff) -- initialized in initStScorePart()
	QString stPid;				// ID
//	QString stPmb;				// MIDI bank
	QString stPmc;				// MIDI channel
	QString stPmp;				// MIDI program
	QString stPnm;				// name
};

#endif /* MUSICXMLIMPORT_H */
